import asyncio
from dataclasses import dataclass
from typing import Any

from redis.asyncio import Redis

from jararaca.presentation.websocket.websocket_interceptor import (
    BroadcastFunc,
    SendFunc,
    WebSocketConnectionBackend,
)


@dataclass
class BroadcastMessage:
    message: bytes

    def encode(self) -> bytes:
        return self.message

    @staticmethod
    def decode(data: bytes) -> "BroadcastMessage":
        return BroadcastMessage(message=data)

    @staticmethod
    def from_message(message: bytes) -> "BroadcastMessage":
        return BroadcastMessage(message=message)


@dataclass
class SendToRoomsMessage:
    rooms: list[str]
    message: bytes

    def encode(self) -> bytes:
        return b"\x00".join([b",".join([a.encode() for a in self.rooms]), self.message])

    @staticmethod
    def decode(data: bytes) -> "SendToRoomsMessage":
        rooms, message = data.split(b"\x00")
        return SendToRoomsMessage(
            rooms=[a.decode() for a in rooms.split(b",")], message=message
        )

    @staticmethod
    def from_message(rooms: list[str], message: bytes) -> "SendToRoomsMessage":
        return SendToRoomsMessage(rooms=rooms, message=message)


class RedisWebSocketConnectionBackend(WebSocketConnectionBackend):
    def __init__(
        self,
        conn: "Redis[bytes]",
        broadcast_pubsub_channel: str,
        send_pubsub_channel: str,
    ) -> None:

        self.redis = conn
        self.broadcast_pubsub_channel = broadcast_pubsub_channel
        self.send_pubsub_channel = send_pubsub_channel

        self.lock = asyncio.Lock()
        self.tasks: set[asyncio.Task[Any]] = set()

    async def broadcast(self, message: bytes) -> None:
        await self.redis.publish(
            self.broadcast_pubsub_channel,
            BroadcastMessage.from_message(message).encode(),
        )

    async def send(self, rooms: list[str], message: bytes) -> None:
        await self.redis.publish(
            self.send_pubsub_channel,
            SendToRoomsMessage.from_message(rooms, message).encode(),
        )

    def configure(
        self, broadcast: BroadcastFunc, send: SendFunc, shutdown_event: asyncio.Event
    ) -> None:

        broadcast_task = asyncio.get_event_loop().create_task(
            self.consume_broadcast(broadcast, shutdown_event)
        )

        send_task = asyncio.get_event_loop().create_task(
            self.consume_send(send, shutdown_event)
        )

        self.tasks.add(broadcast_task)
        self.tasks.add(send_task)

    async def consume_broadcast(
        self, broadcast: BroadcastFunc, shutdown_event: asyncio.Event
    ) -> None:

        async with self.redis.pubsub() as pubsub:
            await pubsub.subscribe(self.broadcast_pubsub_channel)

            while not shutdown_event.is_set():
                message: dict[str, Any] | None = await pubsub.get_message(
                    ignore_subscribe_messages=True
                )

                if message is None:
                    continue

                broadcast_message = BroadcastMessage.decode(message["data"])

                async with self.lock:
                    task = asyncio.get_event_loop().create_task(
                        broadcast(message=broadcast_message.message)
                    )

                    self.tasks.add(task)

                    task.add_done_callback(self.tasks.discard)

    async def consume_send(self, send: SendFunc, shutdown_event: asyncio.Event) -> None:

        async with self.redis.pubsub() as pubsub:
            await pubsub.subscribe(self.send_pubsub_channel)

            while not shutdown_event.is_set():

                message: dict[str, Any] | None = await pubsub.get_message(
                    ignore_subscribe_messages=True
                )

                if message is None:
                    continue

                send_message = SendToRoomsMessage.decode(message["data"])

                async with self.lock:

                    task = asyncio.get_event_loop().create_task(
                        send(send_message.rooms, send_message.message)
                    )

                    self.tasks.add(task)

                    task.add_done_callback(self.tasks.discard)

    async def shutdown(self) -> None:
        async with self.lock:

            await asyncio.gather(*self.tasks, return_exceptions=True)

            await self.redis.close()



import inspect
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Annotated, Any, Callable, Generator, Type, cast, get_origin

from jararaca.core.providers import ProviderSpec, T, Token
from jararaca.microservice import Microservice


class Container:

    def __init__(self, app: Microservice) -> None:
        self.instances_map: dict[Any, Any] = {}

        self.fill_providers(app.providers)

    def fill_providers(self, providers: list[type[Any] | ProviderSpec]) -> None:
        for provider in providers:
            if isinstance(provider, ProviderSpec):
                if provider.use_value:
                    self.instances_map[provider.provide] = provider.use_value
                elif provider.use_class:
                    self.instantiate(provider.use_class, provider.provide)
                elif provider.use_factory:
                    self.instantiate(provider.use_factory, provider.provide)
            else:
                self.instantiate(provider, provider)

    def instantiate(self, type_: type[Any] | Callable[..., Any], bind_to: Any) -> Any:

        dependencies = self.parse_dependencies(type_)

        evaluated_dependencies = {
            name: self.get_or_instantiate(dependency)
            for name, dependency in dependencies.items()
        }

        self.instances_map[bind_to] = type_(**evaluated_dependencies)

    def parse_dependencies(
        self, provider: type[Any] | Callable[..., Any]
    ) -> dict[str, type[Any]]:
        signature = inspect.signature(provider)

        parameters = signature.parameters

        return {
            name: self.lookup_parameter_type(parameter)
            for name, parameter in parameters.items()
            if parameter.annotation != inspect.Parameter.empty
        }

    def lookup_parameter_type(self, parameter: inspect.Parameter) -> Any:
        if parameter.annotation == inspect.Parameter.empty:
            raise Exception(f"Parameter {parameter.name} has no type annotation")

        # if is a Annotated type, return the second argument

        annotation = parameter.annotation
        if get_origin(annotation) is Annotated:

            if len(annotation.__metadata__) == 0:
                raise Exception(f"Parameter {parameter.name} has no type annotation")

            return annotation.__metadata__[0]

        return parameter.annotation

    def get_or_instantiate(self, token_or_type: Type[T] | Token[T]) -> T:

        item_type: Type[T]
        bind_to: Token[T] | Type[T]

        if isinstance(token_or_type, Token):
            item_type = token_or_type.type_
            bind_to = token_or_type
        else:
            item_type = bind_to = token_or_type

        if token_or_type not in self.instances_map:
            self.instantiate(item_type, bind_to)

        return cast(T, self.instances_map[bind_to])

    def get_by_type(self, token: Type[T]) -> T:
        return self.get_or_instantiate(token)

    def get_by_token(self, token: Token[T]) -> T:
        return self.get_or_instantiate(token)


current_container_ctx = ContextVar[Container]("current_container")


def use_current_container() -> Container:
    return current_container_ctx.get()


@contextmanager
def provide_container(container: Container) -> Generator[None, None, None]:
    token = current_container_ctx.set(container)
    try:
        yield
    finally:
        try:
            current_container_ctx.reset(token)
        except ValueError:
            pass



from contextlib import asynccontextmanager
from typing import AsyncContextManager, AsyncGenerator

from jararaca.di import Container
from jararaca.microservice import Microservice, is_interceptor_with_lifecycle


class AppLifecycle:

    def __init__(self, app: Microservice, container: Container) -> None:
        self.app = app
        self.container = container

    @asynccontextmanager
    async def __call__(self) -> AsyncGenerator[None, None]:
        lifecycle_ctxs: list[AsyncContextManager[None]] = []

        for interceptor in self.app.interceptors:
            if is_interceptor_with_lifecycle(interceptor):
                ctx = interceptor.lifecycle(self.app)
                lifecycle_ctxs.append(ctx)

        print("Initializing interceptors lifecycle")
        for ctx in lifecycle_ctxs:
            await ctx.__aenter__()

        yield

        print("Finalizing interceptors lifecycle")
        for ctx in lifecycle_ctxs:
            await ctx.__aexit__(None, None, None)



from typing import Annotated, Any, Generic, Protocol, Type, TypeVar
from uuid import UUID, uuid4

from fastapi import APIRouter, Depends
from pydantic import BaseModel

from jararaca.persistence.base import (
    CRUDOperations,
    Identifiable,
    IdentifiableEntity,
    QueryOperations,
)
from jararaca.persistence.interceptors.aiosqa_interceptor import use_session

ENTITY_T = TypeVar("ENTITY_T", bound=IdentifiableEntity)
SCHEMA_T = TypeVar("SCHEMA_T", bound=BaseModel)


SIMPLE_FILTER_TCO = TypeVar("SIMPLE_FILTER_TCO", contravariant=True)
COMPLEX_FILTER_TCO = TypeVar("COMPLEX_FILTER_TCO", contravariant=True)


class IGenericRouter(
    Protocol, Generic[SCHEMA_T, SIMPLE_FILTER_TCO, COMPLEX_FILTER_TCO]
):
    async def create_task(self, task: SCHEMA_T) -> Identifiable[SCHEMA_T]: ...

    async def get_task(self, task_id: UUID) -> Identifiable[SCHEMA_T]: ...

    async def simple_query_tasks(
        self, filter: Annotated[SIMPLE_FILTER_TCO, Depends(SIMPLE_FILTER_TCO)]
    ) -> list[Identifiable[SCHEMA_T]]: ...

    async def complex_query_tasks(
        self, filter: COMPLEX_FILTER_TCO
    ) -> list[Identifiable[SCHEMA_T]]: ...

    async def update_task(self, task: SCHEMA_T) -> None: ...

    async def delete_task(self, task_id: UUID) -> None: ...

    def get_router(self) -> APIRouter: ...


SIMPLE_FILTER_T = TypeVar("SIMPLE_FILTER_T")
COMPLEX_FILTER_T = TypeVar("COMPLEX_FILTER_T")


# class GenericRouter(Generic[ENTITY_T, SCHEMA_T, SIMPLE_FILTER_T, COMPLEX_FILTER_T]):


def create_generic_router_class(
    entity_type: Type[ENTITY_T],
    schema_type: Type[SCHEMA_T],
    simple_filter_type: Type[SIMPLE_FILTER_T],
    complex_filter_type: Type[COMPLEX_FILTER_T],
    simpel_filter: QueryOperations[SIMPLE_FILTER_T, ENTITY_T],
    complex_filter: QueryOperations[COMPLEX_FILTER_T, ENTITY_T],
) -> Type[IGenericRouter[SCHEMA_T, SIMPLE_FILTER_T, COMPLEX_FILTER_T]]:

    simple_query_operations = simpel_filter
    complex_query_operations = complex_filter
    crud = CRUDOperations[ENTITY_T](entity_type, use_session)

    async def create_task(self: Any, task: schema_type) -> Identifiable[schema_type]:  # type: ignore[valid-type]
        entity = entity_type.from_basemodel(task)
        entity.id = uuid4()
        await crud.create(entity)
        return entity.to_identifiable(schema_type)

    async def get_task(self: Any, task_id: UUID) -> Identifiable[schema_type]:  # type: ignore[valid-type]
        entity = await crud.get(task_id)
        return entity.to_identifiable(schema_type)

    async def simple_query_tasks(
        self: Any, filter: Annotated[SIMPLE_FILTER_T, Depends(simple_filter_type)]
    ) -> list[Identifiable[schema_type]]:  # type: ignore[valid-type]
        entities = await simple_query_operations.query(filter)
        return [task.to_identifiable(schema_type) for task in entities]

    async def complex_query_tasks(
        self: Any, filter: complex_filter_type  # type: ignore[valid-type]
    ) -> list[Identifiable[schema_type]]:  # type: ignore[valid-type]
        entities = await complex_query_operations.query(filter)
        return [task.to_identifiable(schema_type) for task in entities]

    async def update_task(self: Any, task: schema_type) -> None:  # type: ignore[valid-type]
        entity = entity_type.from_basemodel(task)
        await crud.update_by_id(entity.id, entity)

    async def delete_task(self: Any, task_id: UUID) -> None:
        await crud.delete_by_id(task_id)

    def get_router(self: Any) -> APIRouter:
        router = APIRouter(
            tags=[entity_type.__name__],
        )
        router.add_api_route(
            methods=["POST"],
            path="/",
            endpoint=create_task,
        )

        router.add_api_route(
            methods=["GET"],
            path="/{task_id}",
            endpoint=get_task,
        )

        router.add_api_route(
            methods=["GET"],
            path="/",
            endpoint=simple_query_tasks,
        )

        router.add_api_route(
            methods=["POST"],
            path="/query",
            endpoint=complex_query_tasks,
        )

        router.add_api_route(
            methods=["PUT"],
            path="/",
            endpoint=update_task,
        )

        router.add_api_route(
            methods=["DELETE"],
            path="/",
            endpoint=delete_task,
        )

        return router

    return type(
        f"{entity_type.__name__}Router",
        (),
        {
            "entity_type": entity_type,
            "schema_type": schema_type,
            "crud": CRUDOperations[ENTITY_T](entity_type, use_session),
            "simple_query_operations": simpel_filter,
            "complex_query_operations": complex_filter,
            "create_task": create_task,
            "get_task": get_task,
            "simple_query_tasks": simple_query_tasks,
            "complex_query_tasks": complex_query_tasks,
            "update_task": update_task,
            "delete_task": delete_task,
            "get_router": get_router,
        },
    )



import asyncio
import inspect
import logging
import signal
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import Any, AsyncContextManager, AsyncGenerator, Callable, Type, get_origin

import aio_pika
import uvloop
from pydantic import BaseModel

from jararaca.core.uow import UnitOfWorkContextProvider
from jararaca.di import Container
from jararaca.lifecycle import AppLifecycle
from jararaca.messagebus import Message
from jararaca.messagebus.decorators import MESSAGEBUS_INCOMING_MAP, MessageBusController
from jararaca.microservice import Microservice


@dataclass
class AioPikaWorkerConfig:
    url: str
    queue: str
    exchange: str
    prefetch_count: int


class AioPikaMessage(Message[BaseModel]):

    def __init__(
        self,
        aio_pika_message: aio_pika.abc.AbstractIncomingMessage,
        model_type: Type[BaseModel],
    ):
        self.aio_pika_message = aio_pika_message
        self.model_type = model_type

    def payload(self) -> BaseModel:
        return self.model_type.model_validate_json(self.aio_pika_message.body)

    async def ack(self) -> None:
        await self.aio_pika_message.ack()

    async def reject(self) -> None:
        await self.aio_pika_message.reject()

    async def nack(self) -> None:
        await self.aio_pika_message.nack()


class MessageProcessingLocker:

    def __init__(self) -> None:
        self.messages_lock = asyncio.Lock()
        self.current_processing_messages_set: set[asyncio.Task[Any]] = set()

    @asynccontextmanager
    async def lock_message_task(
        self, task: asyncio.Task[Any]
    ) -> AsyncGenerator[None, Any]:
        async with self.messages_lock:
            self.current_processing_messages_set.add(task)
            try:
                yield
            finally:
                self.current_processing_messages_set.discard(task)

    async def wait_all_messages_processed(self) -> None:
        if len(self.current_processing_messages_set) == 0:
            return

        await asyncio.gather(*self.current_processing_messages_set)


class AioPikaMicroserviceProvider:
    def __init__(
        self,
        config: AioPikaWorkerConfig,
        incoming_map: MESSAGEBUS_INCOMING_MAP,
        uow_context_provider: Callable[..., AsyncContextManager[None]],
        app_lifetime: AppLifecycle,
    ):
        self.config = config
        self.incoming_map = incoming_map
        self.uow_context_provider = uow_context_provider
        self.shutdown_event = asyncio.Event()
        self.lock = asyncio.Lock()
        self.tasks: set[asyncio.Task[Any]] = set()
        self.app_lifetime = app_lifetime

    def start_consumer(self) -> None:

        async def consume() -> None:
            async with self.app_lifetime():
                connection = await aio_pika.connect_robust(self.config.url)

                channel = await connection.channel()

                await channel.set_qos(prefetch_count=self.config.prefetch_count)

                topic_exchange = await channel.declare_exchange(
                    self.config.exchange, type="topic"
                )

                queue = await channel.declare_queue(self.config.queue)

                topics = [*self.incoming_map.keys()]

                for topic in topics:
                    await queue.bind(topic_exchange, routing_key=topic)

                await queue.consume(self.message_consumer)

                await self.shutdown_event.wait()
                print("########Worker shutting down")

                async with self.lock:
                    print("########Stopping task incoming")

                print("##########Waiting for all messages to be processed")
                await asyncio.gather(*self.tasks, return_exceptions=True)

                await channel.close()
                await connection.close()

        def on_shutdown(loop: asyncio.AbstractEventLoop) -> None:
            logging.info("Shutting down")
            self.shutdown_event.set()

        with asyncio.Runner(loop_factory=uvloop.new_event_loop) as runner:
            runner.get_loop().add_signal_handler(
                signal.SIGINT, on_shutdown, runner.get_loop()
            )
            runner.run(consume())

    async def message_consumer(
        self, aio_pika_message: aio_pika.abc.AbstractIncomingMessage
    ) -> None:
        if self.shutdown_event.is_set():
            return

        await self.process_message(aio_pika_message)

    async def process_message(
        self, aio_pika_message: aio_pika.abc.AbstractIncomingMessage
    ) -> None:
        async with self.lock:
            task = asyncio.create_task(self.handle_message(aio_pika_message))
            self.tasks.add(task)
            task.add_done_callback(self.tasks.discard)

    async def handle_message(
        self, aio_pika_message: aio_pika.abc.AbstractIncomingMessage
    ) -> None:

        topic = aio_pika_message.routing_key

        if topic is None:
            logging.warning("No topic found for message")

            return

        handler = self.incoming_map.get(topic)

        if handler is None:
            logging.warning("No handler found for topic '%s'" % topic)
            return

        sig = inspect.signature(handler)

        if len(sig.parameters) != 1:
            logging.warning(
                "Handler for topic '%s' must have exactly one parameter" % topic
            )
            return

        parameter = list(sig.parameters.values())[0]

        param_origin = get_origin(parameter.annotation)

        if param_origin is not Message:
            logging.warning(
                "Handler for topic '%s' must have exactly one parameter of type Message"
                % topic
            )
            return

        if len(parameter.annotation.__args__) != 1:
            logging.warning(
                "Handler for topic '%s' must have exactly one parameter of type Message"
                % topic
            )
            return

        message_type = parameter.annotation.__args__[0]

        if not issubclass(message_type, BaseModel):
            logging.warning(
                "Handler for topic '%s' must have exactly one parameter of type Message[BaseModel]"
                % topic
            )
            return

        builded_message = AioPikaMessage(aio_pika_message, message_type)

        try:
            async with self.uow_context_provider():
                await handler(builded_message)
                await aio_pika_message.ack()
        except:
            await aio_pika_message.nack()


def create_messagebus_worker(app: Microservice, config: AioPikaWorkerConfig) -> None:
    container = Container(app)

    combined_messagebus_incoming_map: MESSAGEBUS_INCOMING_MAP = {}

    uow_context_provider = asynccontextmanager(
        UnitOfWorkContextProvider(app=app, container=container)
    )

    for instance_type in app.controllers:
        controller = MessageBusController.get_messagebus(instance_type)

        if controller is None:
            continue

        instance: Any = container.get_by_type(instance_type)

        factory = controller.get_messagebus_factory()(instance)

        for topic, handler in factory.items():
            if topic in combined_messagebus_incoming_map:
                logging.warning(
                    "Topic '%s' already registered by another controller" % topic
                )
            combined_messagebus_incoming_map[topic] = handler

        factory

    AioPikaMicroserviceProvider(
        config=config,
        incoming_map=combined_messagebus_incoming_map,
        uow_context_provider=uow_context_provider,
        app_lifetime=AppLifecycle(app, container),
    ).start_consumer()



from contextlib import asynccontextmanager
from typing import AsyncGenerator

from jararaca.di import Container, provide_container
from jararaca.microservice import AppInterceptor, Microservice


class ContainerInterceptor(AppInterceptor):

    def __init__(self, container: Container) -> None:
        self.container = container

    @asynccontextmanager
    async def intercept(self) -> AsyncGenerator[None, None]:

        with provide_container(self.container):
            yield None


class UnitOfWorkContextProvider:

    def __init__(self, app: Microservice, container: Container):
        self.app = app
        self.container = container
        self.container_interceptor = ContainerInterceptor(container)

    # TODO: Guarantee that the context is closed whenever an exception is raised
    # TODO: Guarantee a unit of work workflow for the whole request, including all the interceptors
    async def __call__(self) -> AsyncGenerator[None, None]:

        ctxs = [self.container_interceptor.intercept()] + [
            interceptor.intercept() for interceptor in self.app.interceptors
        ]

        for ctx in ctxs:
            await ctx.__aenter__()

        yield None

        for ctx in reversed(ctxs):
            await ctx.__aexit__(None, None, None)



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, AsyncContextManager, Protocol

from jararaca.core.providers import ProviderSpec

if TYPE_CHECKING:
    from typing_extensions import TypeIs


class AppInterceptor(Protocol):

    def intercept(self) -> AsyncContextManager[None]: ...


class AppInterceptorWithLifecycle(Protocol):

    def lifecycle(self, app: "Microservice") -> AsyncContextManager[None]: ...


def is_interceptor_with_lifecycle(
    interceptor: Any,
) -> "TypeIs[AppInterceptorWithLifecycle]":

    return hasattr(interceptor, "lifecycle")


@dataclass
class Microservice:
    providers: list[type[Any] | ProviderSpec] = field(default_factory=list)
    controllers: list[type] = field(default_factory=list)
    interceptors: list[AppInterceptor] = field(default_factory=list)



import inspect
from typing import Any, Callable, TypedDict, TypeVar, cast

from fastapi import APIRouter
from fastapi.exceptions import FastAPIError

DECORATED_FUNC = TypeVar("DECORATED_FUNC", bound=Callable[..., Any])
DECORATED_CLASS = TypeVar("DECORATED_CLASS", bound=Any)


class ControllerOptions(TypedDict): ...


class RestController:
    REST_CONTROLLER_ATTR = "__rest_controller__"

    def __init__(
        self, path: str = "", options: ControllerOptions | None = None
    ) -> None:
        self.path = path
        self.options = options
        self.router_factory: Callable[[Any], APIRouter] | None = None

    def get_router_factory(self) -> Callable[[DECORATED_CLASS], APIRouter]:
        if self.router_factory is None:
            raise Exception("Router factory is not set")
        return self.router_factory

    def __call__(self, cls: type[DECORATED_CLASS]) -> type[DECORATED_CLASS]:

        def router_factory(instance: DECORATED_CLASS) -> APIRouter:
            router = APIRouter(
                prefix=self.path,
                **(self.options or {}),
            )

            members = inspect.getmembers(cls, predicate=inspect.isfunction)

            for name, member in members:

                if (mapping := HttpMapping.get_http_mapping(member)) is not None:

                    try:
                        router.add_api_route(
                            methods=[mapping.method],
                            path=mapping.path,
                            endpoint=getattr(instance, name),
                            **(mapping.options or {}),
                        )
                    except FastAPIError as e:
                        raise Exception(
                            f"Error while adding route {mapping.path}"
                        ) from e

            return router

        self.router_factory = router_factory

        RestController.register(cls, self)

        return cls

    @staticmethod
    def register(cls: type[DECORATED_CLASS], controller: "RestController") -> None:
        setattr(cls, RestController.REST_CONTROLLER_ATTR, controller)

    @staticmethod
    def get_controller(cls: type[DECORATED_CLASS]) -> "RestController | None":
        if not hasattr(cls, RestController.REST_CONTROLLER_ATTR):
            return None

        return cast(RestController, getattr(cls, RestController.REST_CONTROLLER_ATTR))


class Options(TypedDict): ...


class HttpMapping:

    HTTP_MAPPING_ATTR = "__http_mapping__"

    def __init__(self, method: str, path: str, options: Options | None = None) -> None:
        self.method = method
        self.path = path
        self.options = options

    def __call__(self, func: DECORATED_FUNC) -> DECORATED_FUNC:

        HttpMapping.register(func, self)

        return func

    @staticmethod
    def register(func: DECORATED_FUNC, mapping: "HttpMapping") -> None:

        setattr(func, HttpMapping.HTTP_MAPPING_ATTR, mapping)

    @staticmethod
    def get_http_mapping(func: DECORATED_FUNC) -> "HttpMapping | None":

        if not hasattr(func, HttpMapping.HTTP_MAPPING_ATTR):
            return None

        return cast(HttpMapping, getattr(func, HttpMapping.HTTP_MAPPING_ATTR))


class Post(HttpMapping):

    def __init__(self, path: str, options: Options | None = None) -> None:
        super().__init__("POST", path, options)


class Get(HttpMapping):

    def __init__(self, path: str, options: Options | None = None) -> None:
        super().__init__("GET", path, options)


class Put(HttpMapping):

    def __init__(self, path: str, options: Options | None = None) -> None:
        super().__init__("PUT", path, options)


class Delete(HttpMapping):

    def __init__(self, path: str, options: Options | None = None) -> None:
        super().__init__("DELETE", path, options)


class Patch(HttpMapping):

    def __init__(self, path: str, options: Options | None = None) -> None:
        super().__init__("PATCH", path, options)



from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator

from fastapi import Depends, FastAPI
from starlette.types import ASGIApp

from jararaca.core.uow import UnitOfWorkContextProvider
from jararaca.di import Container
from jararaca.lifecycle import AppLifecycle
from jararaca.presentation.decorators import RestController
from jararaca.presentation.http_microservice import HttpMicroservice
from jararaca.presentation.websocket.websocket_interceptor import WebSocketInterceptor


class HttpAppLifecycle:

    def __init__(
        self, lifecycle: AppLifecycle, uow_provider: UnitOfWorkContextProvider
    ) -> None:
        self.lifecycle = lifecycle
        self.uow_provider = uow_provider

    @asynccontextmanager
    async def __call__(self, api: FastAPI) -> AsyncGenerator[None, None]:

        websocket_interceptors = [
            interceptor
            for interceptor in self.lifecycle.app.interceptors
            if isinstance(interceptor, WebSocketInterceptor)
        ]

        for interceptor in websocket_interceptors:
            router = interceptor.get_ws_router(
                self.lifecycle.app, self.lifecycle.container, self.uow_provider
            )

            api.include_router(router)

        async with self.lifecycle():
            yield


def create_http_server(
    http_app: HttpMicroservice,
) -> ASGIApp:

    app = http_app.app
    factory = http_app.factory
    container = Container(app)

    uow_provider = UnitOfWorkContextProvider(app, container)

    fastapi_app = (
        factory(
            HttpAppLifecycle(
                AppLifecycle(app, container),
                uow_provider,
            )
        )
        if factory is not None
        else FastAPI()
    )

    fastapi_app.router.dependencies.append(Depends(uow_provider))

    for controller_t in app.controllers:
        controller = RestController.get_controller(controller_t)

        if controller is None:
            continue

        instance: Any = container.get_by_type(controller_t)

        router = controller.get_router_factory()(instance)

        fastapi_app.include_router(router)

    return fastapi_app



from dataclasses import dataclass
from typing import Callable

from fastapi import FastAPI
from starlette.types import Lifespan

from jararaca.microservice import Microservice


@dataclass
class HttpMicroservice:
    app: Microservice
    factory: Callable[[Lifespan[FastAPI]], FastAPI] | None = None
